# Generated by Django 5.2.5 on 2025-09-30 11:06
from __future__ import annotations

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0023_publisher_scope"),
    ]

    operations = [
        migrations.AlterField(
            model_name="profile",
            name="id",
            field=models.CharField(max_length=100, primary_key=True, serialize=False),
        ),
    ]
